#!/usr/bin/python3
import sys
import os
from subprocess import call, DEVNULL, Popen, PIPE
from datetime import datetime
from audit_common import AuditError, read_audit, next_event, next_msg,\
    check_event, assert_audit_enabled, enable_audit_rules, disable_audit_rules


def get_useraud(user):
    cmd = 'useraud ' + user
    p = Popen(cmd, shell=True, stdout=PIPE, stderr=DEVNULL)
    out, err = p.communicate()
    if p.returncode:
        raise AuditError(err.decode())
    out = out.decode()
    return out.split() if out.startswith(user) else []


def useraud(options, user):
    rc = call([
        'useraud',
        user,
        options],
        stdout=DEVNULL)
    if rc:
        raise AuditError("Can't set audit flags to {}.".format(user))


def regtest(names, parts=1):
    print("subtest - {}".format('/'.join(names)))
    for name in names:
        for i in range(parts):
            to_check_messages.append(
                "searching for {}{} events...".format(
                    name,
                    '' if parts == 1 else "({})".format(i + 1)))


# ***************************************************************
try:
    assert_audit_enabled()
    # remove old audit rule
    disable_audit_rules(verbose=False)
except (AuditError, FileNotFoundError) as e:
    print(str(e))
    if os.getuid():
        print("\nPlease, run as root.\n")
    sys.exit(1)

try:
    TESTUSER = 'audittestuser'
    TMPFILE = '/tmp/file-{}'.format(os.getpid())
    saved_root_useraud = ''
    to_check = []
    to_check_messages = []
    enable_audit_rules()
    start = datetime.now()

    print("prepare for test...")
    print("add testuser and set audit flags")
    call(['useradd', '-s', '/bin/false', '-m', TESTUSER], stdout=DEVNULL)
    print("test audit for user :{}".format(TESTUSER))
    call(['passwd', '-d', TESTUSER], stdout=DEVNULL)
    call(["addgroup", TESTUSER, "astra-admin"], stdout=DEVNULL)
    useraud('ocxudnarhwepm', TESTUSER)
    saved_root_useraud = get_useraud('root')
    print("done")

# subtest - uid (i)
    regtest(["uid"])
    useraud('i', 'root')
    call(['su', '-c', 'su -s /bin/sh -- {} -c true'.format(TESTUSER)])
    to_check.append([
        {'type': 'SYSCALL',
         'syscall': 'setuid'}])

# subtest - gid (g)
    regtest(["gid"])
    useraud('g', 'root')
    call(['su', '-c', 'su -s /bin/sh -- {} -c true'.format(TESTUSER)])
    to_check.append([
        {'type': 'SYSCALL',
         'syscall': 'setgid'}])

# subtest - module (l)
    regtest(["init_module", "delete_module"])
    useraud('l', 'root')
    call(['su', '-s', '/bin/sh', '-c', 'modprobe xt_SECMARK'])
    to_check.append([
        # {'type': 'KERN_MODULE',
        #  'name': 'xt_SECMARK'},
        {'type': 'SYSCALL',
         'syscall': 'init_module'}])
    call(['su', '-c', 'rmmod xt_SECMARK'])
    to_check.append([
        # {'type': 'KERN_MODULE',
        #  'name': 'xt_SECMARK'},
        {'type': 'SYSCALL',
         'syscall': 'delete_module'}])

#subtest - create file (c)
    regtest(["create file"])
    call(['su', '-s', '/bin/sh', TESTUSER, '-c', 'touch {}'.format(TMPFILE)])
    to_check.append([
        {'type': 'PATH',
         'name': TMPFILE,
         'nametype': 'CREATE'},
        {'type': 'SYSCALL',
         'syscall': 'open'}])

#subtest -  mac(m)
    regtest(["mac"], 2)
    useraud('m', 'root')
    call(['su', '-c', '/usr/sbin/pdpl-file 255:1:fffffffffffffff:0 {}'.format(TMPFILE)],
         stderr=DEVNULL)
    call(['su', '-c', '/usr/sbin/pdpl-file 0:0:0:0 {}'.format(TMPFILE)])
    to_check.append([
        {'type': 'PATH',
         'name': TMPFILE},
        {'type': 'AVC',
         'parsec': 'ed',  # granted or denied
         'fn': 'sys_chlbl',
         'object': '255:1:fffffffffffffff:0'}])
    to_check.append([
        {'type': 'PATH',
         'name': TMPFILE},
        {'type': 'AVC',
         'parsec': 'ed',  # granted or denied
         'fn': 'sys_chlbl',
         'object': '0:0:0:0'}])

#subtest - open file (o)
    regtest(["open file"])
    call(['su', '-s', '/bin/sh', TESTUSER, '-c', 'cat {}'.format(TMPFILE)])
    to_check.append([
        {'type': 'PATH',
         'name': TMPFILE},
        {'type': 'SYSCALL',
         'syscall': 'open'}])

#subtest - exec (x)
    regtest(["exec program"])
    call(['su', '-s', '/bin/sh', TESTUSER, '-c', '/bin/date'],
         stdout=DEVNULL)
    to_check.append([
        {'type': 'PATH',
         'name': '/bin/date'},
        {'type': 'SYSCALL',
         'syscall': 'exec'}])

#subtest - chmod (d)
    regtest(["chmod"])
    call(['su', '-s', '/bin/sh', TESTUSER, '-c', 'chmod a=rwx {}'.format(TMPFILE)])
    to_check.append([
        {'type': 'PATH',
         'name': TMPFILE},
        {'type': 'SYSCALL',
         'syscall': 'chmod'}])

#subtest - chown (n)
    regtest(["chown"])
    call(['su', '-s', '/bin/sh', TESTUSER, '-c', 'chown root {}'.format(TMPFILE)],
         stderr=DEVNULL)
    to_check.append([
        {'type': 'PATH',
         'name': TMPFILE},
        {'type': 'SYSCALL',
         'syscall': 'chown'}])

#subtest - net (w)
    regtest(["net"])
    call(['su', '-s', '/bin/sh', TESTUSER, '-c', 'ping -c2 localhost'],
         stdout=DEVNULL)
    to_check.append([
        {'type': 'SYSCALL',
         'syscall': 'connect',
         'comm': 'ping'}])

#subtest - chroot (h)
    regtest(["chroot"])
    call(['su', '-s', '/bin/sh', TESTUSER, '-c', '/usr/sbin/chroot /mnt'],
         stderr=DEVNULL)
    to_check.append([
        {'type': 'PATH',
         'name': '/mnt'},
        {'type': 'SYSCALL',
         'syscall': 'chroot'}])

#subtest - rename (e)
    regtest(["rename"])
    call(['su', '-s', '/bin/sh', TESTUSER,
        '-c', 'cp {} {}2 ; mv {}2 {}3'.format(
        TMPFILE, TMPFILE, TMPFILE, TMPFILE
    )])
    to_check.append([
        {'type': 'PATH',
         'name': TMPFILE + '3'},
        {'type': 'PATH',
         'name': TMPFILE + '2'},
        {'type': 'SYSCALL',
         'syscall': 'rename'}])

#subtest - cap (p)
    regtest(["capabilities"])
    call(['usercaps', '-f', TESTUSER],
         stdout=DEVNULL)
    call(['su', '-s', '/bin/sh', TESTUSER, '-c', 'pscaps {} 0 0 0'.format(os.getpid())])
    to_check.append([
        {'type': 'AVC',
         'parsec': 'ed',  # granted or denied
         'fn': 'sys_setcap'}])

#subtest -  audit (a)
    regtest(["audit"])
    call(['su', '-s', '/bin/sh', TESTUSER,
        '-c', '/usr/sbin/setfaud -m o:o:o {}'.format(TMPFILE)])
    to_check.append([
        {'type': 'PATH',
         'name': TMPFILE},
        {'type': 'AVC',
         'parsec': 'ed',  # granted or denied
         'fn': 'ch_aud'}])

#subtest - acl(r)
    regtest(["acl"])
    call(['su', '-s', '/bin/sh', TESTUSER,
        '-c', '/usr/bin/setfacl -m u:root:rwx {}'.format(TMPFILE)])
    to_check.append([
        {'type': 'PATH',
         'name': TMPFILE},
        {'type': 'AVC',
         'parsec': 'ed',  # granted or denied
         'fn': 'acl'}])

#subtest - delete (u)
    regtest(["delete file"])
    call(['su', '-s', '/bin/sh', TESTUSER, '-c', 'rm {}'.format(TMPFILE)])
    to_check.append([
        {'type': 'PATH',
         'name': TMPFILE},
        {'type': 'SYSCALL',
         'syscall': 'unlink'}])

#subtest -  mount (t)
    regtest(["mount"])
    useraud('t', 'root')
    call(['su', '-c', 'mount /home /mnt'],
         stderr=DEVNULL)
    to_check.append([
        {'type': 'PATH',
         'name': '/home'},
        {'type': 'PATH',
         'name': '/mnt'},
        {'type': 'SYSCALL',
         'syscall': 'mount'}])

#subtest -  user
    regtest(["user"], 4)
    audit_user = '/usr/lib/parsec/tests/audit_user'
    call([audit_user])
    to_check.append([
        {'type': 'USER_AVC',
         'exe': audit_user,
         'user_parsec': 'success',
         'eid': '1',
         'msg0': 'audit_user',
         'msg1': 'string0'}])
    to_check.append([
        {'type': 'USER_AVC',
         'exe': audit_user,
         'user_parsec': 'fail',
         'eid': '1',
         'msg0': 'audit_user',
         'msg1': 'string1'}])
    to_check.append([
        {'type': 'USER_AVC',
         'exe': audit_user,
         'user_parsec': 'success',
         'eid': '2',
         'msg0': 'audit_user',
         'msg1': 'string2'}])
    to_check.append([
        {'type': 'USER_AVC',
         'exe': audit_user,
         'user_parsec': 'success',
         'eid': '1',
         'msg0': 'string3'}])

# subtest - search for events
    end = datetime.now()
    disable_audit_rules()

    i = 0
    print(to_check_messages[i])
    for event in next_event(read_audit(start, end)):
        if check_event(to_check[i], event):
            i += 1
            if i == len(to_check):
                sys.exit(0)
            print(to_check_messages[i])

    # print(read_audit(start, end))
    print("Incorrect work of audit system")

except AuditError as e:
    print(str(e))

finally:
    try:
        os.unlink(TMPFILE)
    except FileNotFoundError:
        pass
    try:
        os.unlink(TMPFILE + '3')
    except FileNotFoundError:
        pass
    call(['usercaps', '-d', TESTUSER])
    call(['deluser', '-q' ,'--remove-home', TESTUSER])
    if saved_root_useraud:
        call(['useraud', *saved_root_useraud], stdout=DEVNULL)
    else:
        call(['useraud', '-d', 'root'])

sys.exit(1)
