#!/bin/bash

if [ $(id -u) != 0 ]; then
	>&2 echo "uid != 0, please run this script as root (sudo)"
	exit 1
fi

export GLOBALRES=0
user=cap_tester
bin_name=./cap_extra
cat_loc=$(which cat)
bash_loc=$(which bash)
touch_loc=$(which touch)
ls_loc=$(which ls)

clean_cap()
{
  for cap in "$@"; do setcap -r $cap &>/dev/null; done
}

env_check()
{
	if [ ! $(id -u $user &>/dev/null) ]; then
		useradd -m $user -s /bin/bash &>/dev/null
		echo "$1:1" | chpasswd &>/dev/null
	fi
}

env_clean()
{
	if $(id -u $user &>/dev/null); then
		userdel -r $user &>/dev/null
	fi
}

check_status()
{
	if [ $1 == "err" ]; then
		if [ $2 != 0 ]; then
			echo "$3: PASS"
		else
			>&2 echo "$3: FAIL (!!!)"
			GLOBALRES=1
		fi
	fi

	if [ $1 == "zero" ]; then
		if [ $2 == 0 ]; then
			echo "$3: PASS"
		else
			>&2 echo "$3: FAIL (!!!)"
			GLOBALRES=1
		fi
	fi
}

run_back_proc()
{
	sleep 10&
	sleep 1
	sleep_pid=$(pidof sleep)
}

create_rand_file()
{
	rand_name=$(cat /dev/urandom | tr -dc 'a-z0-9' | head -c 8)
	touch $1/$rand_name
}

test_audit_manage()
{
	sudo -u $user $audit_loc -e0 &>/dev/null
	check_status $1 $? "cap_audit_control"

	sudo -u $user $audit_loc -w /etc/passwd -p wa -k passwdaccess &>/dev/null
	check_status $1 $? "cap_audit_control"

	sudo -u $user $audit_loc -s &>/dev/null
	check_status $1 $? "cap_audit_control"

	sudo -u $user $audit_loc -l &>/dev/null
	check_status $1 $? "cap_audit_control"
}

test_dac_override()
{
	sudo -u $user bash -c 'cd /root/5727' &>/dev/null
	check_status $1 $? "cap_dac_override"

	sudo -u $user $touch_loc /root/5727/aaa &>/dev/null
	check_status $1 $? "cap_dac_override"

	sudo -u $user $ls_loc /root/5727/aaa &>/dev/null
	check_status $1 $? "cap_dac_override"
}

test_mknod()
{
	sudo -u $user $mknod_loc /tmp/null c 4 64 &>/dev/null
	check_status $1 $? "cap_mknod"

	setcap cap_mknod+ep $mknod_loc
	sudo -u $user $mknod_loc /tmp/null c 4 64 &>/dev/null
	check_status $2 $? "cap_mknod"
}

cap_net_raw()
{
	ping_loc=$(which ping)
	clean_cap $ping_loc

	sudo -u $user $ping_loc localhost &>/dev/null
	check_status "err" $? "cap_net_raw"

	setcap cap_net_raw+ep $ping_loc
	sudo -u $user $ping_loc -c 2 localhost &>/dev/null
	check_status "zero" $? "cap_net_raw"

	# ping have default cap_net_raw+ep
	# don't use 'clean_cap' for it
}

cap_kill()
{
	kill_loc=$(which kill)
	clean_cap $kill_loc
	run_back_proc

	sudo -u $user $kill_loc -TERM $sleep_pid &>/dev/null
	check_status "err" $? "cap_kill"

	setcap cap_kill+ep $kill_loc
	sudo -u $user $kill_loc -TERM $sleep_pid &>/dev/null
	check_status "zero" $? "cap_kill"
	clean_cap $kill_loc
}

cap_chown()
{
	chown_loc=$(which chown)
	clean_cap $chown_loc

	create_rand_file "/tmp"
	$chown_loc root: /tmp/$rand_name

	sudo -u $user $chown_loc $user: /tmp/$rand_name &>/dev/null
	check_status "err" $? "cap_chown"

	setcap cap_chown+ep $chown_loc
	sudo -u $user $chown_loc $user: /tmp/$rand_name &>/dev/null
	check_status "zero" $? "cap_chown"

	clean_cap $chown_loc
	rm /tmp/$rand_name
}

cap_setuid()
{
	clean_cap $bin_name
	sudo -u $user $bin_name --uidset
	check_status "err" $? "cap_setuid"

	setcap cap_setuid+ep $bin_name
	sudo -u $user $bin_name --uidset
	check_status "zero" $? "cap_setuid"
	clean_cap $bin_name
}

cap_setgid()
{
	clean_cap $bin_name
	sudo -u $user $bin_name --gidset
	check_status "err" $? "cap_setgid"

	setcap cap_setgid+ep $bin_name
	sudo -u $user $bin_name --gidset
	check_status "zero" $? "cap_setgid"
	clean_cap $bin_name
}

cap_sys_nice()
{
	renice_loc=$(which renice)
	clean_cap $renice_loc
	run_back_proc

	sudo -u $user $renice_loc -n -20 $sleep_pid &>/dev/null
	check_status "err" $? "cap_sys_nice"

	setcap cap_sys_nice+ep $renice_loc
	sudo -u $user $renice_loc -n -20 $sleep_pid &>/dev/null
	check_status "zero" $? "cap_sys_nice"
	kill -TERM $sleep_pid &>/dev/null
	clean_cap $renice_loc
}

cap_fowner()
{
	chmod_loc=$(which chmod)
	clean_cap $chmod_loc
	create_rand_file "/tmp"

	sudo -u $user $chmod_loc 400 /tmp/$rand_name &>/dev/null
	check_status "err" $? "cap_fowner"

	setcap cap_fowner+ep $chmod_loc
	sudo -u $user $chmod_loc 400 /tmp/$rand_name &>/dev/null
	check_status "zero" $? "cap_fowner"

	clean_cap $chmod_loc
	rm /tmp/$rand_name
}

cap_sys_time()
{
	date_loc=$(which date)
	clean_cap $date_loc
	vbox_active=0

	if systemctl is-active --quiet vboxadd-service.service; then
		systemctl stop vboxadd-service.service
		vbox_active=1
	else
		save_date=$(date '+%Y-%m-%d %H:%M:%S')
	fi

	sudo -u $user $date_loc --set="2 OCT 2006 18:00:00" &>/dev/null
	check_status "err" $? "cap_sys_time"

	setcap cap_sys_time+ep $date_loc
	sudo -u $user $date_loc --set="2 OCT 2006 18:00:00" &>/dev/null
	check_status "zero" $? "cap_sys_time"
	clean_cap $date_loc

	# vboxadd-service can auto time sync
	# if not have this service return the recent time
	(($vbox_active)) && systemctl start vboxadd-service.service || date -s "$save_date" &>/dev/null
}

cap_sys_pacct()
{
	clean_cap $bin_name
	create_rand_file "/tmp"
	chown $user: /tmp/$rand_name

	sudo -u $user $bin_name --acct /tmp/$rand_name
	check_status "err" $? "cap_sys_pacct"

	setcap cap_sys_pacct+ep $bin_name
	sudo -u $user $bin_name --acct /tmp/$rand_name
	check_status "zero" $? "cap_sys_pacct"
	clean_cap $bin_name
	rm /tmp/$rand_name
}

cap_ipc_lock()
{
	clean_cap $bin_name
	sudo -u $user $bin_name --mlock
	check_status "err" $? "cap_ipc_lock"

	setcap cap_ipc_lock+ep $bin_name
	sudo -u $user $bin_name --mlock
	check_status "zero" $? "cap_ipc_lock"
	clean_cap $bin_name
}

cap_sys_admin()
{
	clean_cap $bin_name
	curr_hostname=$(hostname)

	sudo -u $user $bin_name --sethostname
	check_status "err" $? "cap_sys_admin"

	setcap cap_sys_admin+ep $bin_name
	sudo -u $user $bin_name --sethostname
	check_status "zero" $? "cap_sys_admin"

	hostname "$curr_hostname" &>/dev/null
	clean_cap $bin_name
}

cap_mknod()
{
	mknod_loc=$(which mknod)
	clean_cap $mknod_loc

	if grep "parsec.strict_mode=" /proc/cmdline &>/dev/null; then
		test_mknod "err" "err"
	else
		test_mknod "err" "zero"
	fi

	clean_cap $mknod_loc
	[ -e /tmp/null ] && rm /tmp/null
}


cap_syslog()
{
	dmesg_loc=$(which dmesg)
	clean_cap $dmesg_loc
	sudo -u $user $dmesg_loc &>/dev/null
	check_status "err" $? "cap_syslog"

	setcap cap_syslog+ep $dmesg_loc
	sudo -u $user $dmesg_loc &>/dev/null
	check_status "zero" $? "cap_syslog"
	clean_cap $dmesg_loc
}

cap_lease()
{
	clean_cap $bin_name
	create_rand_file "/tmp"
	sudo -u $user $bin_name --lease /tmp/$rand_name
	check_status "err" $? "cap_lease"

	setcap cap_lease+ep $bin_name
	sudo -u $user $bin_name --lease /tmp/$rand_name
	check_status "zero" $? "cap_lease"
	clean_cap $bin_name
	rm /tmp/$rand_name
}

cap_wake_alarm()
{
	clean_cap $bin_name
	sudo -u $user $bin_name --wakealarm
	check_status "err" $? "cap_wake_alarm"

	setcap cap_wake_alarm+ep $bin_name
	sudo -u $user $bin_name --wakealarm
	check_status "zero" $? "cap_wake_alarm"
	clean_cap $bin_name
}

cap_sys_rawio()
{
	clean_cap $bin_name
	sudo -u $user $bin_name --iopl
	check_status "err" $? "cap_sys_rawio"

	setcap cap_sys_rawio+ep $bin_name
	sudo -u $user $bin_name --iopl
	check_status "zero" $? "cap_sys_rawio"
	clean_cap $bin_name
}

cap_net_admin()
{
	ifconfig_loc=$(which ifconfig)
	interface=$(ifconfig | awk '{print $1}' | sed -n 1p | tr -d ':')
	clean_cap $ifconfig_loc

	sudo -u $user $ifconfig_loc $interface up &>/dev/null
	check_status "err" $? "cap_net_admin"

	setcap cap_net_admin+ep $ifconfig_loc
	sudo -u $user $ifconfig_loc $interface up &>/dev/null
	check_status "zero" $? "cap_net_admin"
	clean_cap $ifconfig_loc
}

cap_fsetid()
{
	echo_loc=$(which echo)
	clean_cap $echo_loc

	create_rand_file "/tmp"
	chown root: /tmp/$rand_name
	chmod 4777 /tmp/$rand_name

	sudo -u $user $echo_loc write > /tmp/$rand_name
	test "$(ls -l /tmp/$rand_name | grep "rwxrwxrwx")"
	check_status "zero" $? "cap_fsetid"

	setcap cap_fsetid+ep $echo_loc
	chmod 4777 /tmp/$rand_name
	sudo -u $user $echo_loc update >> /tmp/$rand_name
	test "$(ls -l /tmp/$rand_name | grep "rwsrwxrwx")"
	check_status "zero" $? "cap_fsetid"

	clean_cap $echo_loc
	rm /tmp/$rand_name
}

cap_audit_control()
{
	audit_loc=$(which auditctl)
	clean_cap $audit_loc

	test_audit_manage "err"

	setcap cap_audit_control+ep $audit_loc
	test_audit_manage "zero"

	$audit_loc -e1 &>/dev/null
	$audit_loc -W /etc/passwd -p wa -k passwdaccess &>/dev/null
	clean_cap $audit_loc
}

cap_dac_override()
{
	clean_cap $bash_loc $touch_loc $ls_loc
	mkdir /root/5727
	chmod 400 /root/5727

	test_dac_override "err"

	setcap cap_dac_override+ep $bash_loc
	setcap cap_dac_override+ep $(which touch)
	setcap cap_dac_override+ep $ls_loc

	test_dac_override "zero"

	clean_cap $bash_loc $touch_loc $ls_loc
	rm -rf /root/5727
}

cap_dac_read_search()
{
	clean_cap $bash_loc $ls_loc
	mkdir /root/09800

	sudo -u $user bash -c 'cd /root/09800' &>/dev/null
	check_status "err" $? "cap_dac_read_search"

	cd /root/09800
	sudo -u $user $ls_loc /root/09800 &>/dev/null
	check_status "err" $? "cap_dac_read_search"

	setcap cap_dac_read_search+ep $bash_loc
	setcap cap_dac_read_search+ep $ls_loc

	sudo -u $user bash -c 'cd /root/09800' &>/dev/null
	check_status "zero" $? "cap_dac_read_search"

	cd /root/09800
	sudo -u $user $ls_loc /root/09800 &>/dev/null
	check_status "zero" $? "cap_dac_read_search"

	clean_cap $bash_loc $ls_loc
	rm -rf /root/09800
}

cap_linux_immutable()
{
	chattr_loc=$(which chattr)
	clean_cap $chattr_loc

	create_rand_file "/tmp"
	chown $user: /tmp/$rand_name

	# set FS_APPEND_FL flag
	sudo -u $user $chattr_loc +a /tmp/$rand_name &>/dev/null
	check_status "err" $? "cap_linux_immutable"

	setcap cap_linux_immutable+ep $chattr_loc
	sudo -u $user $chattr_loc +a /tmp/$rand_name &>/dev/null
	check_status "zero" $? "cap_linux_immutable"

	# clean 'a' flag from temporary file
	# if have 'a' flag, we can't execute 'rm /tmp/$rand_name'
	sudo -u $user $chattr_loc -a /tmp/$rand_name
	sudo -u $user rm /tmp/$rand_name
	clean_cap $chattr_loc
}

cap_sys_resource()
{
	clean_cap $bin_name
	# increase resource limits
	sudo -u $user $bin_name --rlimit
	check_status "err" $? "cap_sys_resource"

	setcap cap_sys_resource+ep $bin_name
	sudo -u $user $bin_name --rlimit
	check_status "zero" $? "cap_sys_resource"
	clean_cap $bin_name
}

test_pdpl_file()
{
	sudo su - $user bash -c "$pdpl_file_loc /$rand_name &>/dev/null"
	check_status $1 $? "parsec_cap_chmac"

	sudo su - $user bash -c "$pdpl_file_loc 3:0:-1:whole /$rand_name &>/dev/null"
	check_status $2 $? "parsec_cap_chmac"

	sudo su - $user bash -c "$pdpl_file_loc 0:63:0:0 /$rand_name &>/dev/null"
	check_status $3 $? "parsec_cap_chmac"

	pdpl-file 0:63:0:0 /$rand_name
	sudo su - $user bash -c "$pdpl_file_loc /$rand_name &>/dev/null"
	check_status $4 $? "parsec_cap_chmac"

	sudo su - $user bash -c "$pdpl_file_loc 0:0:0:0 /$rand_name &>/dev/null"
	check_status $5 $? "parsec_cap_chmac"
}

parsec_cap_chmac()
{
	usercaps -d $user &>/dev/null
	pdpl_file_loc=/usr/sbin/pdpl-file

	create_rand_file "/"
	test_pdpl_file "zero" "err" "err" "zero" "err"
	rm /$rand_name

	usercaps -m PARSEC_CAP_CHMAC $user &>/dev/null

	create_rand_file "/"
	test_pdpl_file "zero" "zero" "err" "zero" "zero"
	rm /$rand_name

	usercaps -d $user &>/dev/null
}

env_check
echo "Check Linux capabilities work..."
cap_net_raw
cap_kill
cap_chown
cap_setuid
cap_setgid
cap_sys_nice
cap_fowner
cap_sys_time
cap_sys_pacct
cap_ipc_lock
cap_sys_admin
cap_mknod
cap_syslog
cap_lease
if [ $(uname -m) != aarch64 ] && [ $(uname -m) != armv7l ]; then
cap_wake_alarm # ignore ARM -> BT-27678
cap_sys_rawio  # ARM not have sys/io.h prototype
fi
cap_net_admin
if (( ! $(cat /etc/parsec/nochmodx) )); then
cap_fsetid
fi
cap_audit_control
cap_dac_override
cap_linux_immutable
cap_sys_resource
cap_dac_read_search
env_clean

env_check
echo "Check PARSEC capabilities work..."
parsec_cap_chmac
env_clean

exit $GLOBALRES
