#!/bin/bash

set -e # Exit on failure

VERBOSE=
TESTUSER="test-iter-user"		# test user
IMGFILE="/tmp/test-iter-img-file-$$"	# test image file
export GLOBALRES=1	# global result. test is not successful until done
LOOP_DEVICE=

# We do not want to add CCNR at /tmp. Create test FS root at /.
FS_ROOT="/test-iter-fs-root-$$" # test FS root

TMPLOG="/tmp/test-iter-tmp-log-$$"

while getopts "v" OPTION
do
    case $OPTION in
    v) VERBOSE=1 ;;
    esac
done

echo `gettext "# Testing iterate_dir hook"`

cmd() {
    if [ $VERBOSE ]; then
        echo "[#] $*" >&2
    fi
    "$@"
}

fs_up() {
    local fs=$1
    local fs_root=$2
    cmd rm -f $IMGFILE
    cmd truncate $IMGFILE -s 200M
    LOOP_DEVICE=`cmd losetup --show --find $IMGFILE`
    cmd mkfs.$fs $LOOP_DEVICE >/dev/null 2>&1
    cmd mkdir -p $fs_root
    cmd mount $LOOP_DEVICE $fs_root
}

fs_down() {
    cmd umount $LOOP_DEVICE >/dev/null 2>&1 || true
    cmd losetup -d $LOOP_DEVICE >/dev/null 2>&1 || true
    cmd rm -f $IMGFILE
}

create_files() {
    local path=$1
    local n=$2
    local div=$3
    local x=0
    while [ $x -lt $n ]; do
        printf -v fname '%016d' $x
        cmd touch "$path/$fname"
        if [ $(($x %$div )) -eq 0 ]; then
            cmd pdpl-file 1: "$path/$fname"
        fi
        x=$(($x+1))
    done
}

check_dentry_count() {
    local fs=$1
    local path=$2
    local expected_count=$3
    local count=`su -s /bin/sh $TESTUSER -c "ls $path | tee $TMPLOG | wc -l"`
    if [[ $expected_count -ne $count ]]; then
        printf "`gettext "%s: invalid file count (%d, must be %d)\n"`" \
            $fs $count $expected_count >&2
        if [ $VERBOSE ]; then
            echo `gettext "----- ls output -----"` >&2
            cat $TMPLOG >&2
            echo `gettext "--- ls output end ---"` >&2
        fi
        exit $GLOBALRES
    fi
}

add_ccnr() {
    local path=$1
    cmd pdpl-file 1:63:-1:ccnr $path
}

test_hiding_at_fs() {
    local fs=$1
    local rootdir=$2 # CCNR must be set at rootdir

    local path=$rootdir/1
    cmd mkdir -p $path
    add_ccnr $path

    local total=30
    local div=3
    local visible=20

    create_files $path $total $div
    check_dentry_count $fs $path $visible
}

cleanup() {
    fs_down
    cmd rm -f $TMPLOG
    cmd rm -rf $FS_ROOT
    cmd deluser -q --remove-home $TESTUSER
    if [[ $GLOBALRES -ne 0 ]];then
        echo `gettext "FAIL"` >&2
    else
        echo `gettext "PASS"` >&2
    fi
    exit $GLOBALRES
}

cmd trap cleanup EXIT

# create user with ilev==0
cmd useradd -s /bin/false -m $TESTUSER
cmd passwd -d $TESTUSER >/dev/null 2>&1
cmd addgroup $TESTUSER astra-admin

# create FS_ROOT
cmd mkdir -p $FS_ROOT

# test for ext4
fs_up ext4 $FS_ROOT
add_ccnr $FS_ROOT
test_hiding_at_fs ext4 $FS_ROOT
fs_down

# test for xfs
fs_up xfs $FS_ROOT
add_ccnr $FS_ROOT
test_hiding_at_fs xfs $FS_ROOT
fs_down

GLOBALRES=0
